#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/version.h>

#include "lend.h"


#if (LINUX_VERSION_CODE < KERNEL_VERSION(4,17,0))
unsigned long *get_syscall_table(void)
{
	unsigned long *syscall_table;
	unsigned long int i;

	for (i = (unsigned long int)close; i < ULONG_MAX; i += sizeof(void *)) {
		syscall_table = (unsigned long *)i;

		if (syscall_table[__NR_close] == (unsigned long)sys_close)
			return syscall_table;
	}
	return NULL;
}
#else

#define SEARCH_RANGE 512
unsigned char *memmem(const void *haystack, size_t haystacklen, const void *needle, size_t needlelen)
{
    char *p;

    for (p = (char *)haystack; p <= ((char *)haystack - needlelen + haystacklen); p++)
        if (memcmp(p, needle, needlelen) == 0)
            return (void *)p;

    return NULL;
}

static unsigned char *find_do_syscall_64(void) {
	unsigned char *do_syscall_64 = NULL;
	unsigned char *entry_SYSCALL_64 = NULL;

	/* do_syscall_64 has the following signature:
	 *  __visible void do_syscall_64(unsigned long nr, struct pt_regs *regs)
	 *
	 * nr is passed in %rax from userspace, regs is on the top of stack.
	 * So looking for the following pattern should be portable across different
	 * versions of the kernel:
	 *
	 *   48 89 c7                mov    %rax,%rdi
	 *   48 89 e6                mov    %rsp,%rsi
	 *   e8 XX XX XX XX          call   XX XX XX XX
	 */
	unsigned char pattern[] = {0x48, 0x89, 0xc7, 0x48, 0x89, 0xe6, 0xe8};

	// Get address of entry_SYSCALL_64 from LSTAR
	entry_SYSCALL_64 = (unsigned char *)native_read_msr(MSR_LSTAR);

	// The code calls into do_syscall_64: look for the pattern which prepares arguments
	do_syscall_64 = memmem(entry_SYSCALL_64, SEARCH_RANGE, pattern, sizeof(pattern)) + sizeof(pattern);
	if (!do_syscall_64) {
		return NULL;
	}

	// We're pointing to a 32-bit displacement after the fetch phase
	do_syscall_64 = do_syscall_64 + 4 + (int64_t)*((int32_t *)do_syscall_64);

	return do_syscall_64;
}

#ifndef CONFIG_RETPOLINE
static unsigned char *get_table(unsigned char *base)
{
	unsigned char *sys_call_table = NULL;

	/* We look for the actual dispatcher. This is implemented as:
	 *    sys_call_table[nr](regs);
	 * regs is passed to do_syscall_64 as the second parameter and is
	 * generally copied in %rbp, so the following should be a portable
	 * pattern across different versions:
	 *   48 89 ef                mov    %rbp,%rdi
	 *   ff 14 c5 XX XX XX XX    call   *0xXXXXXXXX(,%rax,8)
	 */
	unsigned char pattern[] = {0x48, 0x89, 0xef, 0xff, 0x14, 0xc5};

	// The code calls into do_syscall_64: look for the pattern which prepares arguments
	sys_call_table = memmem(base, SEARCH_RANGE, pattern, sizeof(pattern)) + sizeof(pattern);
	if (!sys_call_table) {
		return NULL;
	}

	return sys_call_table;
}
#else

/**
 * The retpoline has this form:
 *
 *      call 2f
 *   1: pause
 *      lfence
 *      jmp 1b
 *   2: mov %rax, (%rsp)
 *      ret
 *
 * The destination of the call is computed and placed in a register.
 * The register to be used is chosen by the compiler, and the corresponding
 * retpoline is called.
 */

unsigned char *get_table(unsigned char *base)
{
	unsigned char *syscall_table = NULL;
	unsigned int size;
	unsigned char *p = base;

	/* We try to find out the first load of the return address for the retpoline
	 * No easy whay here, except for a fast disassmebly of the function.
	 * We look for instructions in the form:
	 *    48 8b 04 dd 00 00 00    mov    0x0(,%rbx,8),%rax
	 *    48 8b 0c f5 00 00 00    mov    0x0(,%rsi,8),%rcx
	 *    4c 8b 1c dd 00 00 00    mov    0x0(,%rbx,8),%r11
	 * In practice, the destination register can be anything. Also the index,
	 * which is the SIB byte. Matching an instruction of 8 bytes allows us to
	 * discard any check on the SIB byte.
	 * To match any 64-bit mov to any destination register, we must see whether
	 * there is a REX prefix (ignoring the most significant bit of the destination
	 * register). The addressing mode which we want has MOD = 00 and R/M = 100.
	 */
	while(p < base + SEARCH_RANGE) {
		size = length_disasm(p, MODE_X64);
		if(size == 8 && (p[0] & 0x48) == 0x48 // Catches both 0x48 and 0x4c
			     && p[1] == 0x8b 
			     && (p[2] & 0xc0) == 0x00 // mod = 00
			     && (p[2] & 0x04) == 0x04) { // r/m == 100
			break;
		}

		p += size;
	}

	if(p < base + SEARCH_RANGE) {
		syscall_table = (unsigned char *)(int64_t)*((int32_t *)(p + 4));
	}
	
	if (!syscall_table) {
		return NULL;
	}
	return syscall_table;
	
}
#endif /* CONFIG_RETPOLINE */

static unsigned char *get_syscall_table(void)
{
	unsigned char *do_syscall_64;
	unsigned char *table = NULL;

	do_syscall_64 = find_do_syscall_64();

	if(do_syscall_64)
		table = get_table(do_syscall_64);

	return table;
}
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(4,17,0)) */

static int __init syscall_table_init(void)
{
	unsigned char *__sys_call_table = get_syscall_table();
	if(__sys_call_table)
		printk(KERN_INFO KBUILD_MODNAME ": syscall table at %px\n", __sys_call_table);
	else
		printk(KERN_INFO KBUILD_MODNAME ": unable to find syscall table");

	return -1;
}

static void __exit syscall_table_cleanup(void)
{
	// This module will never mount!
}


MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Linux System Call Table Discovery");
MODULE_AUTHOR("Alessandro Pellegrini <pellegrini@diag.uniroma1.it>");

module_init(syscall_table_init);
module_exit(syscall_table_cleanup);
